#include "app/SimpleSlicerWindow.h"
#include "core/ArcBall.h"
#include "transfunc/TransferFunction.h"
#include <math.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <stdio.h>
#include <assert.h>

//#define COLOR_CODE_EDGES





PFNGLTEXIMAGE3DPROC glTexImage3D = NULL;

CSimpleSlicerWindow::CSimpleSlicerWindow()
{
	m_nMode = 0;
	m_bDisplayTF = true;
	// init OpenGL extension
#ifdef WIN32
	glTexImage3D             = (PFNGLTEXIMAGE3DPROC) wglGetProcAddress("glTexImage3D");
#else // LINUX
	glTexImage3D             = (PFNGLTEXIMAGE3DPROC) glxGetProcAddress("glTexImage3D");
#endif
	assert(glTexImage3D != NULL);

	m_pVertices[0] = CVector(-1.0,-1.0,-1.0, 1.0,  0.0, 0.0, 0.0);
	m_pVertices[1] = CVector( 1.0,-1.0,-1.0, 1.0,  1.0, 0.0, 0.0);
	m_pVertices[2] = CVector( 1.0, 1.0,-1.0, 1.0,  1.0, 1.0, 0.0);
	m_pVertices[3] = CVector(-1.0, 1.0,-1.0, 1.0,  0.0, 1.0, 0.0);
															
	m_pVertices[4] = CVector(-1.0,-1.0, 1.0, 1.0,  0.0, 0.0, 1.0);
	m_pVertices[5] = CVector( 1.0,-1.0, 1.0, 1.0,  1.0, 0.0, 1.0);
	m_pVertices[6] = CVector( 1.0, 1.0, 1.0, 1.0,  1.0, 1.0, 1.0);
	m_pVertices[7] = CVector(-1.0, 1.0, 1.0, 1.0,  0.0, 1.0, 1.0);
		
	m_pEdges[0]  = Edge(0,1);
	m_pEdges[1]  = Edge(1,2);
	m_pEdges[2]  = Edge(2,3);
	m_pEdges[3]  = Edge(3,0);
	m_pEdges[4]  = Edge(0,4);
	m_pEdges[5]  = Edge(1,5);
	m_pEdges[6]  = Edge(2,6);
	m_pEdges[7]  = Edge(3,7);
	m_pEdges[8]  = Edge(4,5);
	m_pEdges[9]  = Edge(5,6);
	m_pEdges[10] = Edge(6,7);
	m_pEdges[11] = Edge(7,4);

	m_nNumSlices = 512;
	m_nWidth = 0;
	m_nHeight = 0;

	m_pTransferFunction = new CTransferFunction();

}

CSimpleSlicerWindow::~CSimpleSlicerWindow(void)
{
	
}


bool CSimpleSlicerWindow::loadTextures() {
	
	FILE *pFile = fopen(FILENAME,"rb");
	if (NULL == pFile) {
		return false;
	}

	int size = XMAX*YMAX*ZMAX;
	unsigned char *pVolume = new unsigned char[size];
	bool ok = (size == fread(pVolume,sizeof(unsigned char), size,pFile));
	fclose(pFile);

	glGenTextures(3,m_pTextureIds);

	glBindTexture(GL_TEXTURE_3D,m_pTextureIds[0]);

	// set the texture parameters
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage3D(GL_TEXTURE_3D,0,GL_INTENSITY,XMAX,YMAX,ZMAX,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,pVolume);

	delete [] pVolume;

    return true;
}

void
CSimpleSlicerWindow::cgRenderGeometry() {
	
	int pViewport[4];
	glGetIntegerv(GL_VIEWPORT,pViewport);

	m_nWidth = pViewport[2];
	m_nHeight = pViewport[3];

	static bool bInit = false;

	if (false == bInit) {
		bInit = loadTextures();
		bInit &= createPrograms();
	}

	int i;
	static int nSlice = 0;
	
	const int m_pEdgeList[8][12] = {
		{ 0,1,5,6,   4,8,11,9,  3,7,2,10 }, // v0 is front
		{ 0,4,3,11,  1,2,6,7,   5,9,8,10 }, // v1 is front
		{ 1,5,0,8,   2,3,7,4,   6,10,9,11}, // v2 is front
		{ 7,11,10,8, 2,6,1,9,   3,0,4,5  }, // v3 is front
		{ 8,5,9,1,   11,10,7,6, 4,3,0,2  }, // v4 is front
		{ 9,6,10,2,  8,11,4,7,  5,0,1,3  }, // v5 is front
		{ 9,8,5,4,   6,1,2,0,   10,7,11,3}, // v6 is front
		{ 10,9,6,5,  7,2,3,1,   11,4,8,0 }  // v7 is front
	}; 

#ifdef COLOR_CODE_EDGES
	double colors[12][3] = {
		{1.0,0.0,0.0},{1.0,0.0,0.0},{1.0,1.0,0.0},{1.0,0.0,0.0},
		{0.0,1.0,0.0},{0.0,1.0,0.0},{0.0,1.0,1.0},{0.0,1.0,0.0},
		{0.0,0.0,1.0},{0.0,0.0,1.0},{1.0,0.0,1.0},{0.0,0.0,1.0}
	};
#endif

	float pMatrix[16];

	glGetFloatv(GL_MODELVIEW_MATRIX,pMatrix);

	glDisable(GL_LIGHTING);

	CVector colKeyLight(0.8,0.8,0.8,0.0);
	CVector colFillLight(0.6,0.6,0.5,0.0);
	CVector colBackLight(0.3,0.3,0.2,0.0);

	CMatrix matLight = m_pLightArcball->getRotatonMatrix();
	CVector vecKeyLight  = matLight * CVector(1.0,1.0,1.0,0.0);
	CVector vecFillLight = matLight * CVector(-1.0,1.0,1.0,0.0);
	CVector vecBackLight = matLight * CVector(0.0,1.0,-1.0,0.0);

	
	CVector viewVec(-pMatrix[2],-pMatrix[6],-pMatrix[10],0.0);

	double dMaxDist = viewVec * m_pVertices[0];
	double dMinDist = dMaxDist;
	int nMaxIdx = 0;
	for(int i = 1; i < 8; ++i) {
		double dist = viewVec * m_pVertices[i];
		if ( dist > dMaxDist) {
			dMaxDist = dist;
			nMaxIdx = i;
		}
		if ( dist < dMinDist) {
			dMinDist = dist;
		}

	}

	glEnd();
	
	glLineWidth(2.0);
	glBegin(GL_LINES);
	{
#ifndef COLOR_CODE_EDGES
		glColor3f(1.0,1.0,1.0);
#endif
		
		for(int i = 0; i < 12; i++) {
#ifdef COLOR_CODE_EDGES
			glColor3dv(colors[i]);
#endif
			glVertex4dv(&(m_pVertices[m_pEdges[m_pEdgeList[nMaxIdx][i]].nV1])[0]);
			glVertex4dv(&(m_pVertices[m_pEdges[m_pEdgeList[nMaxIdx][i]].nV2])[0]);
		}

	}
	glEnd();

	
	CGprofile vertProfile = CGlutCgWindow::vertProfile();
	CGprofile fragProfile = CGlutCgWindow::fragProfile();


	CGprogram vertProg; 
	CGprogram fragProg; 

	switch(m_nMode) {

		default:
		case 0:
			vertProg = m_pVertexPrograms[0];
			fragProg = m_pFragmentPrograms[0];
			break;

		case 1:
			vertProg = m_pVertexPrograms[0];
			fragProg = m_pFragmentPrograms[1];
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"viewVec"),&(viewVec[0]));
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"vecKeyLight"),&(vecKeyLight[0]));
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"vecFillLight"),&(vecFillLight[0]));
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"vecBackLight"),&(vecBackLight[0]));
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"colKeyLight"),&(colKeyLight[0]));
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"colFillLight"),&(colFillLight[0]));
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"colBackLight"),&(colBackLight[0]));
			cgGLSetParameter3d(cgGetNamedParameter(fragProg,"k_diff"),0.4,0.2,0.3);
			cgGLSetParameter3d(cgGetNamedParameter(fragProg,"k_spec"),0.4,0.4,0.4);
			cgGLSetParameter3d(cgGetNamedParameter(fragProg,"k_emit"),0.3,0.3,0.3);
			cgGLSetParameter1d(cgGetNamedParameter(fragProg,"shininess"),300.0);
			break;
		case 2:
			vertProg = m_pVertexPrograms[0];
			fragProg = m_pFragmentPrograms[2];
			cgGLSetParameter3dv(cgGetNamedParameter(fragProg,"viewVec"),&(viewVec[0]));
			break;

	}

	
	// Bind uniform parameters to vertex shader
    cgGLSetStateMatrixParameter(cgGetNamedParameter(vertProg, "ModelViewProj"),
                                CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                CG_GL_MATRIX_IDENTITY);

	cgGLSetTextureParameter(   cgGetNamedParameter(fragProg,"volume"),m_pTextureIds[0]);
	cgGLEnableTextureParameter(cgGetNamedParameter(fragProg,"volume"));


	m_pTransferFunction->activate(fragProg);
	
	cgGLEnableProfile(vertProfile);
	cgGLEnableProfile(fragProfile);
    cgGLBindProgram(vertProg);
    cgGLBindProgram(fragProg);
	

	CVector vecStart[12];
	CVector vecDir[12];
	float lambda[12];
	float lambda_inc[12];
	double denom;

	dMinDist += EPSILON;
	dMaxDist -= EPSILON;
	double dPlaneDist    =  dMinDist;   
	double dPlaneDistInc =  (dMaxDist-dMinDist)/double(m_nNumSlices);

	float dVertices[12*3];
	float dEdges[12*3];

	for(i = 0; i < 12; i++) {
		vecStart[i] = m_pVertices[m_pEdges[m_pEdgeList[nMaxIdx][i]].nV1];
		vecDir[i]   = m_pVertices[m_pEdges[m_pEdgeList[nMaxIdx][i]].nV2] - m_pVertices[m_pEdges[m_pEdgeList[nMaxIdx][i]].nV1];

		denom = vecDir[i] * viewVec;

		if (1.0 + denom != 1.0) {
			lambda_inc[i] =  dPlaneDistInc/denom;
			lambda[i]     = (dPlaneDist - vecStart[i] *viewVec)/denom;
		} else {
			lambda[i]     = -1.0;
			lambda_inc[i] =  0.0;	
		}

		dVertices[3*i]   = vecStart[i][0];
		dVertices[3*i+1] = vecStart[i][1];
		dVertices[3*i+2] = vecStart[i][2];
		dEdges[3*i]      = vecDir[i][0];
		dEdges[3*i+1]    = vecDir[i][1];
		dEdges[3*i+2]    = vecDir[i][2];
	};

	CVector intersection[6];
	float lmb[12];

	for(int n = m_nNumSlices-1; n >= 0; --n) {

		for(int e = 0; e < 12; e++) {
			lmb[e] = lambda[e] + n*lambda_inc[e];
		}

		if      ((lmb[0] >= 0.0) && (lmb[0] < 1.0)) intersection[0] = vecStart[0] + lmb[0] * vecDir[0];
		else if ((lmb[1] >= 0.0) && (lmb[1] < 1.0)) intersection[0] = vecStart[1] + lmb[1] * vecDir[1];
		else if ((lmb[3] >= 0.0) && (lmb[3] < 1.0)) intersection[0] = vecStart[3] + lmb[3] * vecDir[3];
		else continue;
		
		if	    ((lmb[2] >= 0.0) && (lmb[2] < 1.0)) intersection[1] = vecStart[2] + lmb[2] * vecDir[2];
		else if ((lmb[0] >= 0.0) && (lmb[0] < 1.0)) intersection[1] = vecStart[0] + lmb[0] * vecDir[0];
		else if ((lmb[1] >= 0.0) && (lmb[1] < 1.0)) intersection[1] = vecStart[1] + lmb[1] * vecDir[1];
		else intersection[1] = vecStart[3] + lmb[3] * vecDir[3];

		if      ((lmb[4] >= 0.0) && (lmb[4] < 1.0)) intersection[2] = vecStart[4] + lmb[4] * vecDir[4];
		else if ((lmb[5] >= 0.0) && (lmb[5] < 1.0)) intersection[2] = vecStart[5] + lmb[5] * vecDir[5];
		else intersection[2] = vecStart[7] + lmb[7] * vecDir[7];
		
		if	    ((lmb[6] >= 0.0) && (lmb[6] < 1.0)) intersection[3] = vecStart[6] + lmb[6] * vecDir[6];
		else if ((lmb[4] >= 0.0) && (lmb[4] < 1.0)) intersection[3] = vecStart[4] + lmb[4] * vecDir[4];
		else if ((lmb[5] >= 0.0) && (lmb[5] < 1.0)) intersection[3] = vecStart[5] + lmb[5] * vecDir[5];
		else intersection[3] = vecStart[7] + lmb[7] * vecDir[7];

		if	    ((lmb[8] >= 0.0) && (lmb[8] < 1.0)) intersection[4] = vecStart[8] + lmb[8] * vecDir[8] ;
		else if ((lmb[9] >= 0.0) && (lmb[9] < 1.0)) intersection[4] = vecStart[9] + lmb[9] * vecDir[9] ;
		else intersection[4] = vecStart[11]+ lmb[11]* vecDir[11];
		
		if	    ((lmb[10]>= 0.0) && (lmb[10]< 1.0)) intersection[5] = vecStart[10]+ lmb[10]* vecDir[10];
		else if ((lmb[8] >= 0.0) && (lmb[8] < 1.0)) intersection[5] = vecStart[8] + lmb[8] * vecDir[8] ;
		else if ((lmb[9] >= 0.0) && (lmb[9] < 1.0)) intersection[5] = vecStart[9] + lmb[9] * vecDir[9] ;
		else intersection[5] = vecStart[11]+ lmb[11]* vecDir[11];

		for(int i = 0; i < 6; ++i) {
			double dLength = intersection[i].getLength();
		}

		//glEnable(GL_TEXTURE_3D);
		//glBindTexture(GL_TEXTURE_3D, m_pTextureIds[0]);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		//glAlphaFunc(GL_GREATER,0.2);
		//glEnable(GL_ALPHA_TEST);
		
		float c = 1.0f-float(n)/float(m_nNumSlices);
		glColor3f(c,c,c);
		glBegin(GL_TRIANGLE_FAN);
		for(int i = 0; i < 6; ++i) {
			intersection[i].glVertex(false,true);
		}
		glEnd();

		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_3D);
	}

	
	cgGLDisableProfile(vertProfile);
	cgGLDisableProfile(fragProfile);
	if (m_bDisplayTF) {
		m_pTransferFunction->render();
	}
	
}


bool CSimpleSlicerWindow::handleButtonEvent(int button, int state, int x, int y) {

	if (m_bDisplayTF) {
		bool b =   m_pTransferFunction->handleButtonEvent(button, state, x, m_nHeight-1-y);
		if (b) glutPostRedisplay();
		return b;
	} 
	return false;
}

bool CSimpleSlicerWindow::handleMoveEvent(int x, int y) {
	if (m_bDisplayTF) {
		bool b =  m_pTransferFunction->handleMoveEvent(x,m_nHeight-1-y);
		if (b) glutPostRedisplay();
		return b;
	} 
	return false;
}
void 
CSimpleSlicerWindow::keyEvent(unsigned char key,int x,int y){

	switch (key) {
		case 'l':
			{
				m_nInteractionMode = MOVE_LIGHT;
			}
			break;
		case 'c':
			{
				m_nInteractionMode = MOVE_CAMERA;
			}
			break;
		case 'm':
			{
				m_nMode++;
			}
			break;
				
		case 'M':
			{
				m_nMode--;
			}
			break;
		case 't':
		case 'T':
			{
				m_bDisplayTF = !m_bDisplayTF;
			}
			break;
		case 'r':
		case 'R':
			{
				m_pCameraArcball->reset();
			}
			break;
	}

	glutPostRedisplay();
}

#define NUM_FRAGPROGS 3
#define NUM_VERTPROGS 1

bool CSimpleSlicerWindow::createPrograms()
{
	CGcontext context = CGlutCgWindow::context();
	CGprofile vertProfile = CGlutCgWindow::vertProfile();
	CGprofile fragProfile = CGlutCgWindow::fragProfile();
    assert(cgIsContext(context));

	m_pVertexPrograms   = new CGprogram[NUM_VERTPROGS];
	m_pFragmentPrograms = new CGprogram[NUM_FRAGPROGS];

	const char *strVertProgFiles[NUM_VERTPROGS] = {"..\\shader\\vp_simple.cg"};
	const char *strFragProgFiles[NUM_FRAGPROGS] = {
		"..\\shader\\fp_emission_absorption.cg", 
		"..\\shader\\fp_emis_abs_gradlightPhong.cg",
		"..\\shader\\fp_showGradient.cg"};
	
	cgGLEnableProfile(vertProfile);
	cgGLEnableProfile(fragProfile);

	bool bOK = true;
	for(int i = 0; i < NUM_VERTPROGS; ++i) {
		m_pVertexPrograms[i] = cgCreateProgramFromFile(context, CG_SOURCE, strVertProgFiles[i], vertProfile, NULL, NULL);
		if (NULL == m_pVertexPrograms[i]) {
			bOK = false;
		}

		if (!cgIsProgramCompiled(m_pVertexPrograms[i])){
			cgCompileProgram(m_pVertexPrograms[i]);
		}
		
		cgGLLoadProgram(m_pVertexPrograms[i]);
	}  
    
	for(int i = 0; i < NUM_FRAGPROGS; ++i) {
		m_pFragmentPrograms[i] = cgCreateProgramFromFile(context, CG_SOURCE, strFragProgFiles[i], fragProfile, NULL, NULL);
		if (NULL == m_pFragmentPrograms[i]) {
			bOK = false;
		}

		if (!cgIsProgramCompiled(m_pFragmentPrograms[i])){
			cgCompileProgram(m_pFragmentPrograms[i]);
		}

		cgGLLoadProgram(m_pFragmentPrograms[i]);
	}  

	return bOK;
}
